from os import listdir
from os.path import isfile, join
from AutoRPE.UtilsRPE.Functions import calls_to_routines, calls_to_functions, \
    is_call_to_function, is_pointer_assignment, is_keyword_argument, is_a_number, \
    find_called_routine, find_call_arguments, find_called_function, \
    type_of_contents, retrieve_variable, in_which_block, remove_if_condition,\
    match_argument_types_and_interface
from AutoRPE.UtilsRPE.Classes import ReadSourceFile, Function, Interface
import warnings
import networkx as nx


def track_dependencies(path_to_model_sources, vault, graph):
    # Find files in sources
    all_sources = [f for f in listdir(path_to_model_sources) if
                   isfile(join(path_to_model_sources, f)) and f.count(".f90")]
    # Process Sources!
    total = len(all_sources)
    for index, filename in enumerate(all_sources):
        print("\rTracing dependencies %i/%i  %20s" % (index + 1, total, filename), end="")
        track_dependencies_in_file(join(path_to_model_sources, filename), vault, graph)


def create_nodes(variables, graph):
    for v in variables:
        try:
            graph[var_identifier(v)]
        except KeyError:
            dim = v.dimension if v.dimension is not None else 0
            graph.add_node(var_identifier(v),
                           dimension=dim,
                           routine=str(v.routine.name),
                           module=str(v.routine.module.name),
                           intent=str(v.intent),
                           type=str(v.type))


def create_edges(argument, dummy, graph):
    v1, v2 = argument, dummy
    dim = v1.dimension if v1.dimension is not None else 0
    graph.add_edge(var_identifier(v1), var_identifier(v2),
                   routine=str(v2.routine.name),
                   kind=str(v2.intent),
                   weight=1+dim/4.
                   )


def track_dependencies_in_file(filename, vault, graph):
    try:
        current_block = vault[filename.split("/")[-1].replace(".f90", "")].main()
    except AttributeError:
        warnings.warn("No module %s" % filename)
        return
    f = ReadSourceFile(filename)

    for line in f.lines:

        # To store calls
        procedure_calls = []

        # To store called procedures names
        called_procedures = []

        # Search for calls to subroutines
        routine_calls = calls_to_routines(line)
        function_calls = calls_to_functions(line, vault)
        if routine_calls:
            for routine_call in routine_calls:
                called_routine = find_called_routine(routine_call)
                if called_routine == "date_and_time":
                    continue
                procedure_calls.append(routine_call)
                called_procedures.append(called_routine)
        # Search for calls to functions
        elif function_calls:
            for function_call in function_calls:
                procedure_calls.append(function_call)
                called_procedures.append(find_called_function(function_call))
        # Look if the code is entering or exiting a routine block
        else:
            current_block = in_which_block(line, current_block, vault=vault)

        # If are there calls to procedures
        if called_procedures:
            for procedure_call, called_procedure in zip(procedure_calls, called_procedures):
                called_arguments = find_call_arguments(procedure_call)
                if not called_arguments:
                    continue

                procedure = vault.get_procedure(called_procedure)
                indices_to_ignore = []

                if isinstance(procedure, Interface):
                    argument_types = [type_of_contents(argument, routine=current_block, vault=vault) for argument in
                                      called_arguments]
                    keyword_arguments = [_x.split("=")[0].strip() if is_keyword_argument(_x) else False for _x in
                                         called_arguments]
                    # Checking if some of the arguments does not necessarily need to be bounded to the dummy
                    # because there's already an existing version within the same interface with a different precision
                    for _in, _type in enumerate(argument_types):
                        new_argument_types = argument_types[:]
                        if _type == "single":
                            new_argument_types[_in] = "double"
                        elif _type == "double":
                            new_argument_types[_in] = "single"
                        else:
                            continue
                        try:
                            match_argument_types_and_interface(new_argument_types, procedure,
                                                               keyword_arguments=keyword_arguments,
                                                               current_block=current_block, vault=vault)
                            indices_to_ignore.append(_in)
                        except AssertionError:
                            continue
                    # Classical match
                    procedure = match_argument_types_and_interface(argument_types, procedure,
                                                                   keyword_arguments=keyword_arguments,
                                                                   current_block=current_block, vault=vault)
                if procedure is None:
                    continue
                for index, argument in enumerate(called_arguments):
                    if index in indices_to_ignore:
                        continue
                    key = None
                    if is_keyword_argument(argument):
                        [key, argument] = argument.split("=", 1)
                        key = key.strip()
                        argument = argument.strip()
                        var_type = type_of_contents(argument, routine=current_block, vault=vault)
                    else:
                        var_type = type_of_contents(argument, routine=current_block, vault=vault)

                    if var_type not in ["single", "double", "rpe_var"] or is_a_number(argument):
                        continue
                    try:
                        var = retrieve_variable(argument, current_block, vault=vault)
                    except AssertionError:
                        continue

                    if key is not None:
                        dummy_var = procedure.get_variable_by_name(key)
                    else:
                        dummy_var = procedure.get_variable_by_position(index)

                    # Graph related stuff
                    # Create nodes
                    create_nodes([var, dummy_var], graph)
                    # Add edges
                    create_edges(var, dummy_var, graph)

        # If there is a pointer assignment on the line
        if is_pointer_assignment(line):
            try:
                # Ignore the if condition. This is not an ideal solution, but the easier for the moment
                line = remove_if_condition(line)
                if line.count("::"):
                    line = line.split("::")[1]
                pointers = line.strip().split("=>")
                ptr = retrieve_variable(pointers[0], current_block, vault=vault)
                target = retrieve_variable(pointers[1], current_block, vault=vault)
            except AssertionError:
                # case in which the pointer is assigned to NULL()
                continue
            if ptr.type != target.type:
                raise AssertionError("Pointer and target should have the same type, something went wrong.")
            if ptr.type not in ["single", "double", "rpe_var"]:
                continue
            # Graph related stuff
            # Create nodes
            create_nodes([ptr, target], graph)
            # Add edges
            create_edges(ptr, target, graph)


def propagate_dependencies(vault, graph=None):
    variables_in_graph = [v for v in vault.variables if var_identifier(v) in graph.nodes]
    var_dict = {}
    for v in variables_in_graph:
        var_dict[var_identifier(v)] = v
    undirected_graph = graph.to_undirected()
    for v in variables_in_graph:
        cc = nx.node_connected_component(undirected_graph, var_identifier(v))
        v.same_as = [var_dict[n] for n in cc]


def var_identifier(var):
    return "%s#%s#%s#%s" % (var.name, var.routine.name, var.routine.module.name, str(var.id))
